from __future__ import annotations

from collections import abc as collections_abc
from decimal import Decimal
from typing import TYPE_CHECKING, AsyncIterator, Generic, TypeVar, cast, overload

from boto3.dynamodb.conditions import ConditionBase, Key
from boto3.exceptions import Boto3Error
from pydantic import BaseModel


TItem = TypeVar("TItem")
if TYPE_CHECKING:
    from types_aiobotocore_dynamodb.client import Exceptions
    from types_aiobotocore_dynamodb.service_resource import (
        DynamoDBServiceResource, Table as _ServiceTable)
    from types_aiobotocore_dynamodb.type_defs import TableAttributeValueTypeDef

    TModel = TypeVar("TModel", bound=BaseModel)
    TItemDict = dict[str, TableAttributeValueTypeDef]

    class AspenExceptions(Exceptions):
        NoSuchKey: type["NoSuchKey"]


class NoSuchKey(Boto3Error):
    pass


def _coerce_decimal_types(value):
    if isinstance(value, Decimal):
        value = float(value)
        return int(value) if value.is_integer() else value
    if isinstance(value, dict):
        return {k: _coerce_decimal_types(v) for k, v in value.items()}
    if isinstance(value, list):
        return [_coerce_decimal_types(v) for v in value]
    if isinstance(value, set):
        return {_coerce_decimal_types(v) for v in value}
    return value


def _sanitize_empty_sets(value):
    if isinstance(value, collections_abc.Set):
        return value if value else []
    if isinstance(value, collections_abc.Mapping):
        return {k: _sanitize_empty_sets(v) for k, v in value.items()}
    if isinstance(value, (list, tuple)):
        return [_sanitize_empty_sets(item) for item in value]
    return value


class DynamoDBTable(Generic[TItem]):
    _table_resource: _ServiceTable | None = None

    @overload
    def __init__(
        self: "DynamoDBTable[TModel]",
        name: str, primary_key: str | tuple[str] | tuple[str, str],
        *, resource: DynamoDBServiceResource, model: type[TModel],
    ) -> None: ...

    @overload
    def __init__(
        self: "DynamoDBTable[TItemDict]",
        name: str, primary_key: str | tuple[str] | tuple[str, str],
        *, resource: DynamoDBServiceResource, model: None = None,
    ) -> None: ...

    def __init__(
        self, name: str, primary_key: str | tuple[str] | tuple[str, str],
        *, resource: DynamoDBServiceResource, model: type[BaseModel] | None = None,
    ):
        if not isinstance(primary_key, tuple):
            primary_key = (primary_key,)
        self.table_name = name
        self.primary_key = primary_key

        self.resource = resource
        self.model = model
        # Shortcut
        self.exceptions = cast("AspenExceptions", self.resource.meta.client.exceptions)
        self.exceptions.NoSuchKey = NoSuchKey

    async def table_resource(self):
        if self._table_resource is None:
            self._table_resource = await self.resource.Table(self.table_name)
        return self._table_resource

    def coerce_item(self, item: dict) -> TItem:
        if self.model:
            return cast(TItem, self.model.model_validate(item))
        return cast(TItem, _coerce_decimal_types(item))

    def key_from_values(self, key_values):
        return {name: value for name, value in zip(self.primary_key, key_values)}

    async def get_item(self, *key_values, **kwargs) -> TItem:
        table = await self.table_resource()

        resp = await table.get_item(Key=self.key_from_values(key_values), **kwargs)
        if "Item" not in resp:
            raise self.exceptions.NoSuchKey(self.table_name, key_values)
        return self.coerce_item(resp["Item"])

    async def put_item(
        self, item: dict | BaseModel, exclude_defaults=False,
        sanitize_empty_sets=True, **kwargs
    ) -> TItem | None:
        table = await self.table_resource()
        if isinstance(item, BaseModel):
            item = item.model_dump(exclude_defaults=exclude_defaults)

        if sanitize_empty_sets:
            item = cast(dict, _sanitize_empty_sets(item))

        resp = await table.put_item(Item=item, **kwargs)
        if "Attributes" in resp:
            return self.coerce_item(resp["Attributes"])

    async def update_item(self, *key_values, **kwargs) -> TItem | None:
        table = await self.table_resource()

        resp = await table.update_item(Key=self.key_from_values(key_values), **kwargs)
        if "Attributes" in resp:
            return self.coerce_item(resp["Attributes"])

    async def delete_item(self, *key_values, **kwargs) -> TItem | None:
        table = await self.table_resource()

        resp = await table.delete_item(Key=self.key_from_values(key_values), **kwargs)
        if "Attributes" in resp:
            return self.coerce_item(resp["Attributes"])

    async def query(self, hash_key, **kwargs) -> tuple[list[TItem], dict | None]:
        table = await self.table_resource()

        if not isinstance(hash_key, ConditionBase):
            hash_key = Key(self.primary_key[0]).eq(hash_key)
        if kwargs.get("ExclusiveStartKey") is None:
            kwargs.pop("ExclusiveStartKey", None)
        resp = await table.query(KeyConditionExpression=hash_key, **kwargs)

        items = [self.coerce_item(item) for item in resp.get("Items", [])]
        return items, resp.get("LastEvaluatedKey", None)

    async def query_all(self, hash_key, **kwargs) -> AsyncIterator[TItem]:
        assert "ExclusiveStartKey" not in kwargs
        kwargs.setdefault("Limit", 50)

        last_key, first_run = None, True
        while last_key or first_run:
            first_run = False
            items, last_key = await self.query(
                hash_key, ExclusiveStartKey=last_key, **kwargs)
            for item in items:
                yield item

    async def scan(self, **kwargs) -> tuple[list[TItem], dict | None]:
        table = await self.table_resource()

        if kwargs.get("ExclusiveStartKey") is None:
            kwargs.pop("ExclusiveStartKey", None)
        resp = await table.scan(**kwargs)

        items = [self.coerce_item(item) for item in resp.get("Items", [])]
        return items, resp.get("LastEvaluatedKey", None)

    async def scan_all(self, **kwargs) -> AsyncIterator[TItem]:
        assert "ExclusiveStartKey" not in kwargs
        kwargs.setdefault("Limit", 50)

        last_key, first_run = None, True
        while last_key or first_run:
            first_run = False
            items, last_key = await self.scan(
                ExclusiveStartKey=last_key, **kwargs)
            for item in items:
                yield item



import argparse
import asyncio
import ssl
import time
from typing import Any, Awaitable, cast

import certifi
import uvloop
from aioboto3 import Session
from aiobotocore.credentials import AioCredentials
from aiodynamo.client import Client
from aiodynamo.credentials import Key as AioDynamoKey, StaticCredentials
from aiodynamo.http.aiohttp import AIOHTTP
from aiodynamo.http.httpx import HTTPX
from aiodynamo.http.types import (
    Request as AioDynamoRequest, Response as AioDynamoResponse)
from aiohttp import ClientSession, TCPConnector
from httpx import AsyncClient as HttpXClient

from aspen_dynamo import DynamoDBTable


boto_session = Session()


async def _get_aiodynamo_cred():
    boto_cred = await cast(Awaitable[AioCredentials], boto_session.get_credentials())
    frozen = await boto_cred.get_frozen_credentials()
    if frozen.access_key and frozen.secret_key:
        return StaticCredentials(
            AioDynamoKey(frozen.access_key, frozen.secret_key, frozen.token)
        )
    raise RuntimeError()


class Runner:
    def __init__(self, table_name: str, table_pk: str, item_id):
        self.table_name = table_name
        self.table_pk = table_pk
        self.item_id = item_id

    async def prepare(self): ...

    async def close(self): ...

    async def get_item(self) -> Any: ...

    async def _capture_naked_aiodynamo_request(self) -> AioDynamoRequest:
        async def capture(req: AioDynamoRequest) -> AioDynamoResponse:
            nonlocal request
            request = req
            raise RuntimeError()
            # return AioDynamoResponse(500, b'')

        request = cast(AioDynamoRequest, None)
        cred = await _get_aiodynamo_cred()
        client = Client(capture, cred, boto_session.region_name)
        try:
            await client.get_item(self.table_name, {self.table_pk: self.item_id})
        except RuntimeError:
            pass
        return request


class AioBoto3ClientRunner(Runner):
    async def prepare(self):
        self._wrapper = boto_session.client("dynamodb")
        self.client = await self._wrapper.__aenter__()

    async def close(self):
        await self._wrapper.__aexit__(None, None, None)

    async def get_item(self):
        return await self.client.get_item(
            TableName=self.table_name,
            Key={self.table_pk: {"N": str(self.item_id)}}
        )


class AioBoto3ResourceRunner(Runner):
    async def prepare(self):
        self._wrapper = boto_session.resource("dynamodb")
        resource = await self._wrapper.__aenter__()
        self.table = await resource.Table(self.table_name)

    async def close(self):
        await self._wrapper.__aexit__(None, None, None)

    async def get_item(self):
        return await self.table.get_item(Key={self.table_pk: self.item_id})


class AioDynamoHttpXRunner(Runner):
    async def prepare(self):
        cred = await _get_aiodynamo_cred()
        client = Client(HTTPX(HttpXClient()), cred, boto_session.region_name)
        self.table = client.table(self.table_name)

    async def get_item(self):
        return await self.table.get_item({self.table_pk: self.item_id})


class AioDynamoAioHttpRunner(Runner):
    async def prepare(self):
        ssl_context = ssl.create_default_context(cafile=certifi.where())
        self._wrapper = ClientSession(connector=TCPConnector(ssl=ssl_context))
        session = await self._wrapper.__aenter__()

        cred = await _get_aiodynamo_cred()
        client = Client(AIOHTTP(session), cred, boto_session.region_name)
        self.table = client.table(self.table_name)

    async def close(self):
        await self._wrapper.__aexit__(None, None, None)

    async def get_item(self):
        return await self.table.get_item({self.table_pk: self.item_id})


class AspenDynamoRunner(Runner):
    async def prepare(self):
        self._wrapper = boto_session.resource("dynamodb")
        resource = await self._wrapper.__aenter__()
        self.table = DynamoDBTable(self.table_name, self.table_pk, resource=resource)

    async def close(self):
        await self._wrapper.__aexit__(None, None, None)

    async def get_item(self):
        return await self.table.get_item(self.item_id)


class NakedHttpXRunner(Runner):
    async def prepare(self):
        self.client = HTTPX(HttpXClient())
        self.request = await self._capture_naked_aiodynamo_request()

    async def get_item(self):
        resp = await self.client(self.request)
        return resp.body.decode()


class NakedAioHttpRunner(Runner):
    async def prepare(self):
        ssl_context = ssl.create_default_context(cafile=certifi.where())
        self._wrapper = ClientSession(connector=TCPConnector(ssl=ssl_context))
        self.client = AIOHTTP(await self._wrapper.__aenter__())

        self.request = await self._capture_naked_aiodynamo_request()

    async def close(self):
        await self._wrapper.__aexit__(None, None, None)

    async def get_item(self):
        resp = await self.client(self.request)
        return resp.body.decode()


async def get_latency(runner: Runner, n=10):
    # Run 10 times and return average run time
    start = time.monotonic()
    for _ in range(n):
        await runner.get_item()
    return (time.monotonic() - start) / n


async def get_load(runner: Runner, n=20, times=5):
    # Run 20 times and return average CPU time
    start = time.process_time()
    for _ in range(times):
        await asyncio.gather(*[
            runner.get_item()
            for _ in range(n)
        ])
    return (time.process_time() - start) / n / times


async def run(table_name: str, table_pk: str, item_id):
    if item_id is None:
        runner = AioBoto3ResourceRunner(table_name, table_pk, None)
        await runner.prepare()
        resp = await runner.table.scan(Limit=1)
        item_id = resp["Items"][0][table_pk]
        print("Item_is is not provided. Found:", item_id)
        await runner.close()

    for runner_class in [
        AioBoto3ClientRunner,
        AioBoto3ResourceRunner,
        AspenDynamoRunner,
        NakedAioHttpRunner,
        AioDynamoAioHttpRunner,
        NakedHttpXRunner,
        AioDynamoHttpXRunner,
    ]:
        runner = runner_class(table_name, table_pk, item_id)

        print(f"Running {runner_class.__name__}...")
        await runner.prepare()
        print("==== resp ", "=" * 40)
        print(await runner.get_item())
        print("=" * 50)

        latency = await get_latency(runner)
        print(f"Latency: {latency * 1000:.2f}ms")

        load = await get_load(runner)
        print(f"CPU time: {load * 1000:.2f}ms")

        await runner.close()
        print()


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Benchmark DynamoDB clients.")
    parser.add_argument("name", help="DynamoDB table name.")
    parser.add_argument("pk", help="Primary key attribute name.")
    parser.add_argument(
        "item_id", nargs="?", type=int, default=None,
        help="Item id (optional, default: 0).")
    args = parser.parse_args()
    uvloop.run(run(args.name, args.pk, args.item_id))
